"""module for setonsave modifier."""
from __future__ import annotations
from typing import Callable, Any, TYPE_CHECKING
from inspect import signature
from .modifier import Modifier
if TYPE_CHECKING:
    from ..ctx import Ctx


class SetOnSaveModifier(Modifier):
    """Setonsave modifier updates or sets value on save."""

    def __init__(self, setter: Callable) -> None:
        if not callable(setter):
            raise ValueError('setonsave setter is not callable')
        params_len = len(signature(setter).parameters)
        if params_len > 1:
            raise ValueError('not a valid setonsave setter')
        self.setter = setter

    def serialize(self, ctx: Ctx) -> Any:
        params_len = len(signature(self.setter).parameters)
        if params_len == 1:
            return self.setter(ctx.val)
        return self.setter()



from typing import Optional, TypedDict
from jsonclasses import jsonclass


class Preference(TypedDict):
    ios: bool
    android: bool


@jsonclass
class AutoSetting:
    user: Optional[str]
    preference: Preference



from __future__ import annotations
from typing import TypedDict
from jsonclasses import jsonclass, jsondict

@jsondict
class GroupedInner(TypedDict):
    on: bool
    off: bool


@jsondict
class Grouped(TypedDict):
    ios: GroupedInner
    android: GroupedInner


@jsonclass
class NestShapeUser:
    name: str
    grouped: Grouped




"""
This module contains `jsondict`, the decorator for typed dicts used with
jsonclasses.
"""
from typing import Optional, Union, Callable, overload
from .cgraph import CGraph


@overload
def jsondict(cls: type) -> type: ...


@overload
def jsondict(
    cls: None,
    class_graph: Optional[str] = 'default'
) -> Callable[[type], type[dict]]: ...


@overload
def jsondict(
    cls: type,
    class_graph: Optional[str] = 'default'
) -> type[dict]: ...


def jsondict(
    cls: Optional[type] = None,
    class_graph: Optional[str] = 'default',
) -> Union[Callable[[type], type[dict]], type[dict]]:
    """The jsondict typed dict class decorator. To declare a jsonclass typed
    dict, use this syntax:

      @jsondict
      class MyDict(TypedDict):
        my_field_one: str
        my_field_two: bool
    """
    if cls is not None:
        if not isinstance(cls, type):
            raise ValueError('@jsondict should be used to decorate a class.')
        cgraph = CGraph(class_graph)
        cgraph.put_dict(cls)
        return cls
    else:
        def parametered_jsondict(cls):
            return jsondict(
                cls,
                class_graph=class_graph)
        return parametered_jsondict



"""
JSON Classes
~~~~~~~~~~~~

JSON Classes is the Modern Declarative Data Flow and Data Graph Framework for
the AI Empowered Generation.

:copyright: (c) 2020 by Fillmula Inc., Victor Teo

:license: MIT, see LICENSE for more details.
"""
# flake8: noqa: F401
from .jsonclass import jsonclass
from .jsondict import jsondict
from .jsonenum import jsonenum
from .types import types
from .typing import linkto, linkedby, linkedthru
from .json_encoder import JSONEncoder
from .isjsonclass import isjsonclass, isjsonobject



from __future__ import annotations
from unittest import TestCase
from jsonclasses.excs import ValidationException
from tests.classes.simple_mixed import SimpleMixed


class TestUnion(TestCase):

    def test_union_raises_if_none_matches(self):
        mixed = SimpleMixed(name='Mixed', mixed=5.0)
        with self.assertRaises(ValidationException):
            mixed.validate()

    def test_union_does_not_raise_if_any_matches(self):
        mixed = SimpleMixed(name='Mixed', mixed=5)
        mixed.validate()
        mixed1 = SimpleMixed(name='Mixed', mixed='5')
        mixed1.validate()



from __future__ import annotations
from unittest import TestCase
from jsonclasses.excs import ValidationException
from tests.classes.simple_config_user import SimpleConfigUser
from tests.classes.simple_folder import SimpleFolder
from tests.classes.simple_node import SimpleNode
from tests.classes.simple_shape_setting import SimpleShapeSetting
from tests.classes.simple_shorthand_setting import SimpleShorthandSetting
from tests.classes.default_shape_value import DefaultShapeValue


class TestShape(TestCase):

    def test_shape_accepts_shorthand_types(self):
        setting = SimpleShapeSetting()
        setting.email['auto_send'] = '5'
        setting.email['receive_promotion'] = '5'
        self.assertRaisesRegex(ValidationException,
                               "'email\\.auto_send': value is not bool",
                               setting.validate)

    def test_shape_accepts_optional_shorthand_types(self):
        setting = SimpleShorthandSetting()
        setting.email['auto_send'] = '5'
        setting.email['receive_promotion'] = '5'
        self.assertRaisesRegex(ValidationException,
                               "'email\\.auto_send': value is not bool",
                               setting.validate)
        setting = SimpleShorthandSetting()
        setting.validate()

    def test_shape_accepts_typed_dict_type(self):
        value = DefaultShapeValue(
            settings={'ios': '2', 'android': '2', 'name': '4'})
        self.assertRaisesRegex(
            ValidationException,
            "'settings\\.ios': value is not bool",
            value.validate)

    def test_shape_validates_inner_fields(self):
        user = SimpleConfigUser(config={})
        self.assertRaisesRegex(ValidationException,
                               '\'config\\.ios\': value required',
                               user.validate)

    def test_shape_doesnt_raise_on_validation_if_inner_fields_are_ok(self):
        user = SimpleConfigUser(config={'ios': True, 'android': True})
        user.validate()

    def test_shape_assigns_none_on_inner_fields_for_accessing(self):
        user = SimpleConfigUser(config={})
        self.assertEqual(user.config, {'ios': None, 'android': None})

    # TODO: make shape work according to strict settings
    def test_shape_sanitizes_input(self):
        user = SimpleConfigUser(config={'haha': True, 'android': False})
        self.assertEqual(user.config, {'ios': None, 'android': False})

    def test_shape_underscores_keys_on_init(self):
        folder = SimpleFolder(
            config={'displaySize': True, 'displayDate': False})
        self.assertEqual(folder.config,
                         {'display_size': True, 'display_date': False})

    def test_shape_doesnt_underscore_keys_on_init_if_specified(self):
        node = SimpleNode(
            config={'displaySize': True, 'display_Date': False})
        self.assertEqual(node.config,
                         {'display_size': None, 'display_date': None})

    def test_shape_camelizes_keys_on_tojson(self):
        folder = SimpleFolder(
            config={'displaySize': True, 'displayDate': False})
        self.assertEqual(folder.tojson()['config'],
                         {'displaySize': True, 'displayDate': False})

    def test_shape_doesnt_camelize_keys_on_tojson(self):
        node = SimpleNode(
            config={'display_size': True, 'display_date': False})
        self.assertEqual(node.tojson()['config'],
                         {'display_size': True, 'display_date': False})

    def test_shape_produce_validation_error_message_for_one_item(self):
        node = SimpleNode(config={})
        with self.assertRaises(ValidationException) as context:
            node.validate()
        exception = context.exception
        self.assertEqual(len(exception.keypath_messages), 1)
        self.assertEqual(exception.keypath_messages['config.display_size'],
                         "value required")

    def test_strict_shape_raises_if_key_is_not_allowed(self):
        with self.assertRaisesRegex(ValidationException,
                                    "'config\\.a': key is not allowed"):
            SimpleFolder(config={'a': True, 'b': False})

    def test_shape_can_be_accessed_with_dot_notation(self):
        folder = SimpleFolder(config={'display_size': True, 'display_date': False})
        self.assertEqual(folder.config.display_size, True)
        self.assertEqual(folder.config.display_date, False)

    def test_shape_can_be_set_with_dot_notation(self):
        folder = SimpleFolder(config={'display_size': True, 'display_date': False})
        folder.config.display_size = False
        folder.config.display_date = True
        self.assertEqual(folder.config['display_size'], False)
        self.assertEqual(folder.config['display_date'], True)



"""This modules contains the JSONClasses types modifier."""
from __future__ import annotations
from typing import Callable, Any, Optional, Union
from datetime import datetime
from copy import deepcopy
from .fdef import Fdef
from .keypath import new_mongoid
from .modifiers import (BoolModifier, ChainedModifier,
                        CompareModifier, DateModifier, DatetimeModifier,
                        DefaultModifier, DictOfModifier, EagerModifier,
                        EmbeddedModifier, EnumModifier,
                        FloatModifier, IndexModifier, InputAllModifier,
                        InputNameModifier, InputLnameModifier,
                        InputValueModifier, InstanceOfModifier,
                        IntModifier, InvalidModifier, LengthModifier,
                        LinkedByModifier, LinkedInModifier,
                        LinkedThruModifier, LinkToModifier, ListOfModifier,
                        MatchModifier, MaxModifier, MaxlengthModifier,
                        LtModifier, MinModifier, MinlengthModifier, GtModifier,
                        NegativeModifier, NonnullModifier, NullableModifier,
                        OneOfModifier, UnionModifier, OnWriteModifier,
                        OnSaveModifier, OnUpdateModifier, OutputLnameModifier,
                        OutputNameModifier, OutputValueModifier,
                        PositiveModifier, PresentModifier,
                        NonnegativeModifier, NonpositiveModifier,
                        PresentWithModifier, PresentWithoutModifier,
                        PreserializeModifier, PrimaryModifier,
                        RangeModifier, ReadonlyModifier, ReadwriteModifier,
                        RefereeModifier, ReferrerModifier,
                        RequiredModifier, ResetModifier, SetOnSaveModifier,
                        ShapeModifier, StrModifier, StrictModifier,
                        TempModifier, TransformModifier, TrimModifier,
                        TruncateModifier, UniqueModifier, ValidateModifier,
                        Modifier, WriteNonnullModifier, WriteonceModifier,
                        WriteonlyModifier, DenyModifier, CascadeModifier,
                        NullifyModifier, OpModifier, AsopModifier,
                        AsopdModifier, UrlModifier, EmailModifier,
                        DigitModifier,AlphaModifier,NumericModifier,
                        AlnumModifier,ToTitleModifier,ToCapModifier,
                        ToLowerModifier,ToUpperModifier,RoundModifier,
                        UnresolvedModifier, AnyModifier)

Str = str
Int = int
Float = float


class Types:
    """The class of types marks object. Types marks provide necessary
    information about an json object's shape, transformation, validation,
    serialization and sanitization.
    """

    def __init__(  # pylint: disable=keyword-arg-before-vararg
        self,
        original: Optional[Types] = None,
        *args: Modifier
    ) -> None:
        if not original:
            self.fdef = Fdef()
            self.modifier = ChainedModifier()
        else:
            self.fdef = deepcopy(original.fdef)
            modifier = original.modifier
            for arg in args:
                modifier = modifier.append(arg)
                arg.define(self.fdef)
            self.modifier = modifier

    @property
    def invalid(self) -> Types:
        """Fields marked with invalid will never be valid, thus these fields
        will never pass validation.
        """
        return Types(self, InvalidModifier())

    @property
    def primary(self) -> Types:
        """Field marked with primary become the object's primary key.
        """
        return Types(self, PrimaryModifier())

    @property
    def readonly(self) -> Types:
        """Fields marked with readonly will not be able to go through
        initialization and set method. You can update value of these fields
        directly or through update method. This prevents client side to post
        data directly into these fields.

        `writeonce`, `readonly` and `writenonnull` cannot be presented
        together.
        """
        return Types(self, ReadonlyModifier())

    @property
    def writeonly(self) -> Types:
        """Fields marked with writeonly will not be available in outgoing json
        form. Users' password is a great example of writeonly.
        """
        return Types(self, WriteonlyModifier())

    @property
    def readwrite(self) -> Types:
        """Fields marked with readwrite will be presented in both inputs and
        outputs. This is the default behavior. And this specifier can be
        omitted.
        """
        return Types(self, ReadwriteModifier())

    @property
    def writeonce(self) -> Types:
        """Fields marked with writeonce can only be set once through
        initialization and set method. You can update value of these fields
        directly or through update method. This is suitable for e.g. dating app
        user gender. Gender should not be changed once set.

        `writeonce`, `readonly` and `writenonnull` cannot be presented
        together.
        """
        return Types(self, WriteonceModifier())

    @property
    def writenonnull(self) -> Types:
        """Fields marked with writenonnull can only be set to a nonnull value.
        The update method doesn't have this limitation. This prevents user from
        setting present value back into None.

        `writeonce`, `readonly` and `writenonnull` cannot be presented
        together.
        """
        return Types(self, WriteNonnullModifier())

    @property
    def internal(self) -> Types:
        """Fields marked with internal will not be accepted as input, and it
        will not be present in output. These fields are internal and hidden
        from users.
        """
        return Types(self, ReadonlyModifier(), WriteonlyModifier())

    @property
    def temp(self) -> Types:
        """Fields marked with temp won't be written into database. As soon as
        database writing happens, value of this field is cleared and set to
        None. Examples of it's use cases are authentication code validation,
        input validation, etc.
        """
        return Types(self, TempModifier())

    @property
    def index(self) -> Types:
        """Fields marked with index are picked up by ORM integrations to setup
        database column index for you. This modifier doesn't have any effect
        around transforming and validating.
        """
        return Types(self, IndexModifier())

    @property
    def unique(self) -> Types:
        """Fields marked with unique are picked up by ORM integrations to setup
        database column unique index for you. This modifier doesn't have any
        effect around transforming and validating. When database engine raises
        an exception, jsonclasses's web framework integration will catch it and
        return 400 automatically.

        If you are implementing jsonclasses ORM integration, you should use
        UniqueFieldException provided by jsonclasses.excs to keep
        consistency with other jsonclasses integrations.
        """
        return Types(self, UniqueModifier())

    @property
    def embedded(self) -> Types:
        """Instance fields marked with the embedded mark is embedded into the
        hosting document for noSQL databases.
        """
        return Types(self, EmbeddedModifier())

    @property
    def linkto(self) -> Types:
        """In a database relationship, fields marked with linkto save an id of
        the object being referenced at the local table.
        """
        return Types(self, LinkToModifier())

    def linkedby(self, foreign_key: str) -> Types:
        """In a database relationship, fields marked with linkedby find
        reference from the destination table.
        """
        return Types(self, LinkedByModifier(foreign_key))

    def linkedthru(self, foreign_key: str) -> Types:
        """In a database relationship, fields marked with linkedthru save
        relationships to a designated association table and find references
        through it.
        """
        return Types(self, LinkedThruModifier(foreign_key))

    def linkedin(self, cls: Any) -> Types:
        """In a database relationship, fields marked with linkedin save
        relationships to the table under provided class.
        """
        return Types(self, LinkedInModifier(cls))

    def referrer(self, referrer_key: str) -> Types:
        """In a many to many database relationship, fields marked with referrer
        has a provided custom key name in the association table.
        """
        return Types(self, ReferrerModifier(referrer_key))

    def referee(self, referee_key: str) -> Types:
        """In a many to many database relationship, fields marked with referee
        reference the other side of the relationship with this provided custom
        key name.
        """
        return Types(self, RefereeModifier(referee_key))

    @property
    def nullify(self) -> Types:
        """When an object is deleted, linked objects' references are set to
        null instead of deleted.
        """
        return Types(self, NullifyModifier())

    @property
    def cascade(self) -> Types:
        """When an object is deleted, linked objects with cascade relationship
        are deleted.
        """
        return Types(self, CascadeModifier())

    @property
    def deny(self) -> Types:
        """When an object is deleted, linked objects with deny relationship
        prevent this object being deleted.
        """
        return Types(self, DenyModifier())

    @property
    def str(self) -> Types:
        """Fields marked with str should be str type. This is a type modifier.
        """
        return Types(self, StrModifier())

    def match(self, pattern: Str) -> Types:
        """Fields marked with match are tested againest the argument regular
        expression pattern.
        """
        return Types(self, MatchModifier(pattern))

    def oneof(self, str_list: list[Str]) -> Types:
        """This is the enum equivalent for jsonclasses. Values in the provided
        list are considered valid values.
        """
        return Types(self, OneOfModifier(str_list))

    def minlength(self, length: int) -> Types:
        """Values at fields marked with minlength should have a length which is
        not less than length.

        Args:
          length (int): The minimum length required for the value.

        Returns:
          Types: A new types chained with this modifier.
        """
        return Types(self, MinlengthModifier(length))

    def maxlength(self, length: int) -> Types:
        """Values at fields marked with maxlength should have a length which is
        not greater than length.

        Args:
          length (int): The minimum length required for the value.

        Returns:
          Types: A new types chained with this modifier.
        """
        return Types(self, MaxlengthModifier(length))

    def length(self, minlength: int, maxlength: Optional[int] = None) -> Types:
        """Fields marked with length should have a length which is between the
        two arguments. If only one argument is provided, the value length
        should be exactly that length.

        Args:
          minlength (int): The minimum length required for the value.
          maxlength (Optional[int]): The maximum length required for the value.

        Returns:
          Types: A new types chained with this modifier.
        """
        return Types(self, LengthModifier(minlength, maxlength))

    @property
    def url(self) -> Types:
        """Fields marked with url should be valid url string.
        """
        return Types(self, UrlModifier())

    @property
    def digit(self) -> Types:
        """Values of fields marked with digit should be valid digit string.
        """
        return Types(self, DigitModifier())

    @property
    def alpha(self) -> Types:
        """Values of fields marked with alpha should be valid alpha string.
        """
        return Types(self, AlphaModifier())

    @property
    def numeric(self) -> Types:
        """Values of fields marked with numeric should be valid numeric string.
        """
        return Types(self, NumericModifier())

    @property
    def email(self) -> Types:
        """Values of fields marked with email should be valid email format.
        """
        return Types(self, EmailModifier())

    @property
    def alnum(self) -> Types:
        """Values fields marked with alnum should be valid alnum strings.
        """
        return Types(self, AlnumModifier())

    @property
    def int(self) -> Types:
        """Fields marked with int should be int type. This is a type modifier.
        """
        return Types(self, IntModifier())

    @property
    def float(self) -> Types:
        """Fields marked with float should be float type. This is a type
        modifier.
        """
        return Types(self, FloatModifier())

    def min(self, value: Float) -> Types:
        """Fields marked with min are tested again this value. Values less than
        the argument value are considered invalid.
        """
        return Types(self, MinModifier(value))

    def max(self, value: Float) -> Types:
        """Fields marked with max are tested again this value. Values greater
        than the argument value are considered invalid.
        """
        return Types(self, MaxModifier(value))

    def lte(self, value: Float) -> Types:
        """Fields marked with lte are tested again this value. Values greater
        than the argument value are considered invalid.
        """
        return Types(self, MaxModifier(value))

    def gte(self, value: Float) -> Types:
        """Fields marked with gte are tested again this value. Values less than
        the argument value are considered invalid.
        """
        return Types(self, MinModifier(value))

    def lt(self, value: Float) -> Types:
        """Fields marked with lt are tested again this value. Values greater
        or equal than the argument value are considered invalid.
        """
        return Types(self, LtModifier(value))

    def gt(self, value: Float) -> Types:
        """Fields marked with gt are tested again this value. Values less than
        or equal than the argument value are considered invalid.
        """
        return Types(self, GtModifier(value))


    def range(self, min_value: Float, max_value: Float) -> Types:
        """Fields marked with range are tested again argument values. Only
        values between the arguments range are considered valid.
        """
        return Types(self, RangeModifier(min_value, max_value))

    @property
    def negative(self) -> Types:
        """Fields marked with negative should have a value less than zero.
        """
        return Types(self, NegativeModifier())

    @property
    def positive(self) -> Types:
        """Fields marked with negative should have a value greater than zero.
        """
        return Types(self, PositiveModifier())

    @property
    def nonnegative(self) -> Types:
        """Fields marked with nonnegative should have a value greater than or
        equal to zero.
        """
        return Types(self, NonnegativeModifier())

    @property
    def nonpositive(self) -> Types:
        """Fields marked with nonnegative should have a value less than or
        equal to zero.
        """
        return Types(self, NonpositiveModifier())

    @property
    def bool(self) -> Types:
        """Fields marked with bool should be bool type. This is a type modifier.
        """
        return Types(self, BoolModifier())

    @property
    def date(self) -> Types:
        """Fields marked with date should be date type. This is a type modifier.
        """
        return Types(self, DateModifier())

    @property
    def datetime(self) -> Types:
        """Fields marked with datetime should be datetime type. This is a type
        modifier.
        """
        return Types(self, DatetimeModifier())

    def enum(self, enum_class: Union[type, str]) -> Types:
        """Fields marked with enum should be enum value of provided enum type.
        This is a type modifier.
        """
        return Types(self, EnumModifier(enum_class))

    @property
    def inputall(self) -> Types:
        """Inputall makes enum field to accept all kinds of acceptable enum
        values in any forms.
        """
        return Types(self, InputAllModifier())

    @property
    def inputlname(self) -> Types:
        """Inputlname makes enum field to accept enum's lowercase name as
        input.
        """
        return Types(self, InputLnameModifier())

    @property
    def inputname(self) -> Types:
        """Inputlname makes enum field to accept enum's uppercase name as
        input.
        """
        return Types(self, InputNameModifier())

    @property
    def inputvalue(self) -> Types:
        """Inputlname makes enum field to accept enum's value as input.
        """
        return Types(self, InputValueModifier())

    @property
    def outputlname(self) -> Types:
        """Outputlname makes enum field to output lowercase name as display
        value.
        """
        return Types(self, OutputLnameModifier())

    @property
    def outputname(self) -> Types:
        """Outputname makes enum field to output uppercase name as display
        value.
        """
        return Types(self, OutputNameModifier())

    @property
    def outputvalue(self) -> Types:
        """Outputvalue makes enum field to output value as display value.
        """
        return Types(self, OutputValueModifier())

    def listof(self, item_types: Any) -> Types:
        """Fields marked with listof should be a list of the given type. This
        is a type modifier.
        """
        return Types(self, ListOfModifier(item_types))

    def dictof(self, item_types: Any) -> Types:
        """Fields marked with listof should be a str keyed dict of the given
        type. This is a type modifier.
        """
        return Types(self, DictOfModifier(item_types))

    def shape(self, item_types_map: dict[Str, Any]) -> Types:
        """Fields marked with shape are objects shaped with given shape. This
        is a type modifier.
        """
        return Types(self, ShapeModifier(item_types_map))

    @property
    def strict(self) -> Types:
        """Shape fields marked with strict disallow undefined keys.
        """
        return Types(self, StrictModifier())

    def instanceof(self, json_object_class: Any) -> Types:
        """Fields marked with instance of are objects of given class.
        """
        return Types(self, InstanceOfModifier(json_object_class))

    def objof(self, jcls: Any) -> Types:
        """Fields marked with objof are objects of given class.
        """
        return Types(self, InstanceOfModifier(jcls))

    def union(self, type_list: list[Any]) -> Types:
        """Fields marked with union accepts value from these types.
        """
        return Types(self, UnionModifier(type_list))

    @property
    def any(self) -> Types:
        """Fields marked with any can be any value.
        """
        return Types(self, AnyModifier())

    @property
    def required(self) -> Types:
        """Fields marked with required are invalid when value is None.

        Returns:
          Types: A new types chained with this modifier.
        """
        return Types(self, RequiredModifier())

    @property
    def nullable(self) -> Types:
        """Fields marked with nullable can be None. This is the default
        behavior even without this modifier. It's the opposite to required
        modifier. Values inside lists have implicitly required modifier. Use
        this to allow null or None values inside lists.

        Returns:
          Types: A new types chained with this modifier.
        """
        return Types(self, NullableModifier())

    @property
    def present(self) -> Types:
        """When validating, field marked with present, can not pass validation
        if it has a None value. This is useful for foreign key fields to do
        required validation.
        """
        return Types(self, PresentModifier())

    def presentwith(self, referring_key: str) -> Types:
        """Fields marked with presentwith modifier are forced presented if
        referring field is present. If referring field has None value, this
        field's value is optional. If referring field has non None value, value
        of this field is required.
        """
        return Types(self, PresentWithModifier(referring_key))

    def presentwithout(self, referring_keys: Union[str, list[str]]) -> Types:
        """Fields marked with presentwithout modifier are forced presented if
        referring field is not present. If referring field has None value, this
        field's value should be present. If referring field has non None value,
        value of this field is not forced to be present.
        """
        return Types(self, PresentWithoutModifier(referring_keys))

    def validate(self, validate_callable: Callable) -> Types:
        """The validate field mark takes a modifier callable as its sole
        argument. Use this to define custom field value validations.

        Args:
            validate_callable (Callable): The validate callable takes up to 2
            arguments, which are value and context. Returning None or True
            means the value is valid, while returning a str message or False
            means validation failed.

        Returns:
            Types: A new types chained with this modifier.
        """
        return Types(self, ValidateModifier(validate_callable))

    def op(self, op_callable: Callable) -> Types:
        """Operator modifier validates value against the operator object user
        passed in. This modifier is special and doesn't bypass None value. If
        the operator is not present, this modifier fails.

        Args:
            op_callable (Callable): The op callable takes 1 to 4 arguments. The
            first is the operator object, the second is the object being
            operated, the third is the value of the field, the fourth is the
            validating context. Returning None or True means the value is
            valid, while returning a str message or False means validation
            failed.

        Returns:
            Types: A new types chained with this modifier.
        """
        return Types(self, OpModifier(op_callable))

    def compare(self, compare_callable: Callable) -> Types:
        """The compare field mark takes a modifier callable as its sole
        argument. If value of compared fields are changed, this modifier is
        called with 2 to 5 arguments.

        Args:
            compare_callable (Callable): The compare callable. Arg 1 is the old
            value, arg 2 is the new value, arg 3 is key path, arg 4 is the
            object, arg 5 is the context. Returning None means the value is
            valid, while returning a str message means validation failed.

        Returns:
            Types: A new types chained with this modifier.
        """
        return Types(self,
                     ResetModifier(),
                     CompareModifier(compare_callable))

    # transformers

    def default(self, value: Any) -> Types:
        """During initialization, if values of fields with default are not
        provided. The default value is used instead of leaving blank.

        Args:
          value (any): The default value of this field. If the value is
          callable, it's return value is used.

        Returns:
          Types: A new types chained with this modifier.
        """
        return Types(self, DefaultModifier(value))

    def truncate(self, max_length: Int) -> Types:
        """During initialization and set, if string value is too long, it's
        truncated to argument max length.

        Args:
          max_length (int): The allowed max length of the field value.

        Returns:
          Types: A new types chained with this modifier.
        """
        return Types(self, EagerModifier(), TruncateModifier(max_length))

    @property
    def trim(self) -> Types:
        """This modifier will trim string value. Remove leading and trailing
        whitespaces.

        Returns:
            Types: A new types chained with this modifier.
        """
        return Types(self, EagerModifier(), TrimModifier())

    @property
    def totitle(self) -> Types:
        """This modifier titlizes string.

        Returns:
            Types: A new types chained with this modifier.
        """
        return Types(self, EagerModifier(), ToTitleModifier())

    @property
    def tocap(self) -> Types:
        """This modifier capitalizes string.

        Returns:
            Types: A new types chained with this modifier.
        """
        return Types(self, EagerModifier(), ToCapModifier())

    @property
    def tolower(self) -> Types:
        """This modifier lowercasefies string.

        Returns:
            Types: A new types chained with this modifier.
        """
        return Types(self, EagerModifier(), ToLowerModifier())

    @property
    def toupper(self) -> Types:
        """This modifier uppercasefies string.

        Returns:
            Types: A new types chained with this modifier.
        """
        return Types(self, EagerModifier(), ToUpperModifier())

    @property
    def round(self) -> Types:
        """This modifier rounds number value.

        Returns:
            Types: A new types chained with this modifier.
        """
        return Types(self, EagerModifier(), RoundModifier())

    def transform(self, transformer: Callable) -> Types:
        """This mark applies transfromer on the value. When value is None, the
        transformer is not called. This class barely means to transform. Use
        default mark with a callable to assign calculated default value.

        Args:
          transformer (Callable): This transformer function takes one argument
          which is the current value of the field.

        Returns:
          Types: A new types chained with this modifier.
        """
        return Types(self, EagerModifier(), TransformModifier(transformer))

    def uploader(self, uploader: Callable) -> Types:
        """Uploader is barely a syntax alias for transformer. Uploaded files
        can be processed through a transformer and get url attached to this
        object.

        Args:
          transformer (Callable): This transformer function takes one argument
          which is the current value of the field.

        Returns:
          Types: A new types chained with this modifier.
        """
        return Types(self, EagerModifier(), TransformModifier(uploader))

    def asop(self, asop_transformer: Callable) -> Types:
        """Asop modifier assigns transformed operator value to this field. When
        the operator is not present, a ValidationException is raised.

        Args:
            asop_transformer (Callable): This transformer function takes 1 to
            3 arguments. The first one is the opeartor, the second one is the
            value of the field, the third is the transforming context.

        Returns:
            Types: A new types chained with this modifier.
        """
        return Types(self, AsopModifier(asop_transformer))

    @property
    def asopd(self) -> Types:
        """Asopd modifier assigns operator value directly to this field without
        any transforming. When the operator  is not present, a
        ValidationException is raised.

        Returns:
            Types: A new types chained with this modifier.
        """
        return Types(self, AsopdModifier())

    def setonsave(self, setter: Callable) -> Types:
        """Setonsave modifier marks a field to be updated just before
        serializing into the database if this field is modified and to be
        serialized.

        Args:
          setter (Callable): This setter function takes zero or one argument
          which is the current value of the field.

        Returns:
          Types: A new types chained with this modifier.
        """
        return Types(self, PreserializeModifier(), SetOnSaveModifier(setter))

    def onsave(self, callback: Callable) -> Types:
        """Onsave inserts a callback into the modifier chain. If save action
        is triggered, the callback is called with the current value.

        Args:
            callback (Callable): This callback function takes one argument
            which is the current value of the field.

        Returns:
            Types: A new types chained with this modifier.
        """
        return Types(self, OnSaveModifier(callback))

    def onupdate(self, callback: Callable) -> Types:
        """Onupdate is a callback modifier. If value updated when saving, this
        callback is called with the current value or both values.

        Args:
            callback (Callable): A callable which takes arguments.

        Returns:
            Types: A new types chained with this modifier.
        """
        return Types(self, ResetModifier(), OnUpdateModifier(callback))

    def onwrite(self, callback: Callable) -> Types:
        """Onwrite is a callback modifier. Whenever a new value is being
        serialized into the database, this is called.

        Args:
            callback (Callable): A callable which takes arguments.

        Returns:
            Types: A new types chained with this modifier.
        """
        return Types(self, OnWriteModifier(callback))

    @property
    def nonnull(self) -> Types:
        """This modifier is a instructional transformer designated for shape,
        dictof and listof. This is not a modifier. To mark a field is required
        and should not be null, use `required` instead. This transformer should
        be used right before shape, dictof and listof, to given an instruction
        of not leaving null for the field.

        Returns:
          Types: A new types chained with this modifier.
        """
        return Types(self, NonnullModifier())

    # compound

    @property
    def mongoid(self: Types) -> Types:
        """This modifier assigns a default bson object id to the field.
        """
        return Types(self, DefaultModifier(lambda: new_mongoid()))

    @property
    def tscreated(self: Types) -> Types:
        """This modifier adds a default current time to the field.
        """
        return Types(self, DefaultModifier(datetime.now))

    @property
    def tsupdated(self: Types) -> Types:
        """This modifier adds a default current time and a set on save.
        """
        return Types(self,
                     DefaultModifier(datetime.now),
                     SetOnSaveModifier(lambda: datetime.now()))

    # internal

    def _unresolved(self: Types, arg: str) -> Types:
        """This modifier marks unresolved status. This is used internally. Do
        not use this.
        """
        return Types(self, UnresolvedModifier(arg))


types = Types()
"""The root of the types modifier. To mark an field with type annotation,
accessor annotation, modifier annotation and transformer annotation, use types
like this:

  @jsonclass
  class MyObject:
    my_field_one: bool = types.bool.readonly.required
    my_field_two: password = types.bool.writeonly.length(8, 16).transform(salt\
).required
"""



from __future__ import annotations
from unittest import TestCase
from jsonclasses.excs import ValidationException
from tests.classes.simple_company import SimpleCompany
from tests.classes.simple_code import SimpleCode


class TestLength(TestCase):

    def test_length_does_not_raise_length_in_between(self):
        c1 = SimpleCompany(name='AZ', number_of_employees=5)
        c1.validate()
        c2 = SimpleCompany(name='QWERT', number_of_employees=5)
        c2.validate()
        c3 = SimpleCompany(name='QWERTQWERT', number_of_employees=5)
        c3.validate()

    def test_length_raises_length_lt_lower(self):
        c1 = SimpleCompany(name='Q', number_of_employees=5)
        with self.assertRaises(ValidationException):
            c1.validate()

    def test_length_raises_length_gt_lower(self):
        c1 = SimpleCompany(name='QWERTYUIOP{', number_of_employees=5)
        with self.assertRaises(ValidationException):
            c1.validate()

    def test_length_should_be_match(self):
        code = SimpleCode(code='1234')
        code.validate()

    def test_length_raises_if_length_not_match(self):
        code = SimpleCode(code='12345')
        with self.assertRaises(ValidationException):
            code.validate()
        code = SimpleCode(code='1')
        with self.assertRaises(ValidationException):
            code.validate()



"""The dot notation accessible dict."""
from __future__ import annotations
from typing import Any
from .owned_dict import OwnedDict


class ObjectDict(OwnedDict):

    def __getattr__(self, name):
        try:
            return super().__getattr__(name)
        except Exception:
            return self[name]

    def __setattr__(self, name: str, value: Any) -> None:
        if name in self.keys():
            self[name] = value
        else:
            super().__setattr__(name, value)



"""module for listof modifier."""
from __future__ import annotations
from typing import Any, Collection, Iterable
from ..fdef import FType
from .collection_type_modifier import CollectionTypeModifier


class ListOfModifier(CollectionTypeModifier):
    """This modifier validates list."""

    def __init__(self, raw_item_types: Any) -> None:
        super().__init__(raw_item_types)
        self.cls = list
        self.ftype = FType.LIST

    def enumerator(self, value: list) -> Iterable:
        return enumerate(value)

    def empty_collection(self) -> Collection:
        return []

    def append_value(self, i: int, v: Any, col: list):
        col.append(v)



"""module for unique modifier."""
from ..fdef import Fdef
from .modifier import Modifier


class UniqueModifier(Modifier):
    """Unique modifier marks a column should be unique."""

    def define(self, fdef: Fdef) -> None:
        fdef._unique = True



"""module for linkedthru modifier."""
from ..fdef import DeleteRule, Fdef, FStore
from .modifier import Modifier


class LinkedThruModifier(Modifier):
    """Linked by modifier marks a field linked with a joinning table."""

    def __init__(self, foreign_key: str) -> None:
        self.foreign_key = foreign_key

    def define(self, fdef: Fdef) -> None:
        fdef._fstore = FStore.FOREIGN_KEY
        fdef._foreign_key = self.foreign_key
        fdef._use_join_table = True
        if fdef._delete_rule is None:
            fdef._delete_rule = DeleteRule.NULLIFY



from typing import TypedDict
from jsonclasses import jsonclass, types


class Settings(TypedDict):
    ios: bool
    android: bool
    name: str


@jsonclass
class DefaultShapeValue:
    settings: Settings = types.shape(Settings) \
        .default({'ios': False, 'android': True, 'name': 'set'})



"""This module defines `JConf`. Each JSON class has its own configuration. The
configuration object tweaks the behavior of JSON classes.
"""
from __future__ import annotations
from typing import Optional, Callable, Any, Union, cast, final, TYPE_CHECKING
from .jobject import JObject
if TYPE_CHECKING:
    from .jfield import JField
    from .cgraph import CGraph


OnCreate = Callable[[JObject, Any], None]
OnUpdate = Callable[[JObject, Any], None]
OnDelete = Callable[[JObject, Any], None]
CanCreate = Callable[[JObject, Any], Union[bool, None, str]]
CanUpdate = Callable[[JObject, Any], Union[bool, None, str]]
CanDelete = Callable[[JObject, Any], Union[bool, None, str]]
CanRead = Callable[[JObject, Any], Union[bool, None, str]]


@final
class JConf:
    """The configuration of JSON classes. Each JSON class has its own
    configuration that each instance shares. This object tweaks the behavior of
    JSON classes.
    """

    def __init__(self: JConf,
                 cgraph: Optional[str],
                 key_encoding_strategy: Optional[Callable[[str], str]],
                 key_decoding_strategy: Optional[Callable[[str], str]],
                 strict_input: Optional[bool],
                 ref_key_encoding_strategy: Optional[Callable[[JField], str]],
                 validate_all_fields: Optional[bool],
                 abstract: Optional[bool],
                 reset_all_fields: Optional[bool],
                 on_create: Optional[Union[OnCreate, list[OnCreate]]],
                 on_update: Optional[Union[OnUpdate, list[OnUpdate]]],
                 on_delete: Optional[Union[OnDelete, list[OnDelete]]],
                 can_create: Optional[Union[CanCreate, list[CanCreate]]],
                 can_update: Optional[Union[CanUpdate, list[CanUpdate]]],
                 can_delete: Optional[Union[CanDelete, list[CanDelete]]],
                 can_read: Optional[Union[CanRead, list[CanRead]]]) -> None:
        """
        Initialize a new configuration object.

        Args:
            cgraph (Optional[str]): The name of the class graph on which \
                the JSON class is defined.
            key_encoding_strategy (Optional[Callable[[str], str]]): How \
                object keys are encoded. The default is camelize.
            key_decoding_strategy (Optional[Callable[[str], str]]): How \
                object keys are decoded. The default is underscore.
            strict_input (Optional[bool]): Whether raise errors on receiving \
                invalid input keys.
            ref_key_encoding_strategy (Optional[Callable[[JField], str]]): The \
                reference field local key conversion function.
            validate_all_fields (Optional[bool]): The default field \
                validating method when performing saving and validating.
            abstract (Optional[bool]): Instance of abstract classes cannot \
                be initialized.
            reset_all_fields (Optional[bool]): Whether record all previous \
                values of an object and enable reset functionality.
            on_create (Optional[Union[OnCreate, list[OnCreate]]]): The callback
                on first time save.
            on_update (Optional[Union[OnUpdate, list[OnUpdate]]]): The callback
                on existing object save.
            on_delete (Optional[Union[OnDelete, list[OnDelete]]]): The callback
                on existing object deletion.
            can_create (Optional[Union[CanCreate, list[CanCreate]]]): The
                creation guard.
            can_update (Optional[Union[CanUpdate, list[CanUpdate]]]): The
                updation guard.
            can_delete (Optional[Union[CanDelete, list[CanDelete]]]): The
                deletion guard.
            can_read (Optional[Union[CanRead, list[CanRead]]]): The reading
                guard.
        """
        self._cls: Optional[type[JObject]] = None
        self._cgraph = cgraph or 'default'
        self._key_encoding_strategy = key_encoding_strategy
        self._key_decoding_strategy = key_decoding_strategy
        self._strict_input = strict_input
        self._ref_key_encoding_strategy = ref_key_encoding_strategy
        self._validate_all_fields = validate_all_fields
        self._abstract = abstract
        self._reset_all_fields = reset_all_fields
        if callable(on_create):
            self._on_create = [on_create]
        elif isinstance(on_create, list):
            self._on_create = on_create
        else:
            self._on_create = []
        if callable(on_update):
            self._on_update = [on_update]
        elif isinstance(on_update, list):
            self._on_update = on_update
        else:
            self._on_update = []
        if callable(on_delete):
            self._on_delete = [on_delete]
        elif isinstance(on_delete, list):
            self._on_delete = on_delete
        else:
            self._on_delete = []
        if callable(can_create):
            self._can_create = [can_create]
        elif isinstance(can_create, list):
            self._can_create = can_create
        else:
            self._can_create = []
        if callable(can_update):
            self._can_update = [can_update]
        elif isinstance(can_update, list):
            self._can_update = can_update
        else:
            self._can_update = []
        if callable(can_delete):
            self._can_delete = [can_delete]
        elif isinstance(can_delete, list):
            self._can_delete = can_delete
        else:
            self._can_delete = []
        if callable(can_read):
            self._can_read = [can_read]
        elif isinstance(can_read, list):
            self._can_read = can_read
        else:
            self._can_read = []

    def __eq__(self: JConf, other: Any) -> bool:
        if not isinstance(other, JConf):
            return False
        other_config = cast(JConf, other)
        if self.cgraph != other_config.cgraph:
            return False
        if self.key_encoding_strategy != other_config.key_encoding_strategy:
            return False
        if self.key_decoding_strategy != other_config.key_decoding_strategy:
            return False
        if self.strict_input != other_config.strict_input:
            return False
        if self.ref_key_encoding_strategy != other_config.ref_key_encoding_strategy:
            return False
        if self.validate_all_fields != other_config.validate_all_fields:
            return False
        if self.abstract != other_config.abstract:
            return False
        if self.reset_all_fields != other_config.reset_all_fields:
            return False
        if self.on_create != other_config.on_create:
            return False
        if self.on_update != other_config.on_update:
            return False
        if self.on_delete != other_config._on_delete:
            return False
        if self.can_create != other_config.can_create:
            return False
        if self.can_update != other_config.can_update:
            return False
        if self.can_delete != other_config.can_delete:
            return False
        if self.can_read != other_config.can_read:
            return False
        return True

    @property
    def cls(self: JConf) -> type[JObject]:
        """The JSON class on which this class config is defined.
        """
        return cast(type[JObject], self._cls)

    @property
    def cgraph(self: JConf) -> CGraph:
        """The name of the class graph on which the JSON class is defined.
        """
        from .cgraph import CGraph
        return CGraph(self._cgraph)

    @property
    def key_encoding_strategy(self: JConf) -> Callable[[str], str]:
        """The object key encoding strategy.
        """
        if self._key_encoding_strategy is None:
            return self.cgraph.default_config.key_encoding_strategy
        return self._key_encoding_strategy

    @property
    def key_decoding_strategy(self: JConf) -> Callable[[str], str]:
        """The object key decoding strategy.
        """
        if self._key_decoding_strategy is None:
            return self.cgraph.default_config.key_decoding_strategy
        return self._key_decoding_strategy

    @property
    def strict_input(self: JConf) -> bool:
        """Whether raise errors on receiving invalid input keys.
        """
        if self._strict_input is None:
            return self.cgraph.default_config.strict_input
        return self._strict_input

    @property
    def ref_key_encoding_strategy(self: JConf) -> Callable[[JField], str]:
        """The reference field local key conversion function.
        """
        if self._ref_key_encoding_strategy is None:
            return self.cgraph.default_config.ref_key_encoding_strategy
        return self._ref_key_encoding_strategy

    @property
    def validate_all_fields(self: JConf) -> bool:
        """The default field validating method when performing saving and
        validating.
        """
        if self._validate_all_fields is None:
            return self.cgraph.default_config.validate_all_fields
        return self._validate_all_fields

    @property
    def abstract(self: JConf) -> bool:
        """Instance of abstract classes cannot be initialized.
        """
        if self._abstract is None:
            return self.cgraph.default_config.abstract
        return self._abstract

    @property
    def reset_all_fields(self: JConf) -> bool:
        """Whether record all previous values of an object and enable reset
        functionality.
        """
        if self._reset_all_fields is None:
            return self.cgraph.default_config.reset_all_fields
        return self._reset_all_fields

    @property
    def on_create(self: JConf) -> list[OnCreate]:
        """The object creation callback.
        """
        return self._on_create

    @property
    def on_update(self: JConf) -> list[OnUpdate]:
        """The object saving callback.
        """
        return self._on_update

    @property
    def on_delete(self: JConf) -> list[OnDelete]:
        """The object deleting callback.
        """
        return self._on_delete

    @property
    def can_create(self: JConf) -> list[CanCreate]:
        """The object creation guard.
        """
        return self._can_create

    @property
    def can_update(self: JConf) -> list[CanUpdate]:
        """The object updation guard.
        """
        return self._can_update

    @property
    def can_delete(self: JConf) -> list[CanDelete]:
        """The object deletion guard.
        """
        return self._can_delete

    @property
    def can_read(self: JConf) -> list[CanRead]:
        """The object reading guard.
        """
        return self._can_read



"""module for shape modifier."""
from __future__ import annotations
from jsonclasses.keypath import concat_keypath
from jsonclasses.vmsgcollector import VMsgCollector
from typing import Any, Sequence, Union, TYPE_CHECKING
from ..fdef import Fdef, FType, Nullability, Strictness
from ..excs import ValidationException
from ..jconf import JConf
from .type_modifier import TypeModifier
if TYPE_CHECKING:
    from ..ctx import Ctx


class ShapeModifier(TypeModifier):
    """Shape modifier validates a dict of values with defined shape."""

    def __init__(self, raw_shape_types: Union[dict[str, Any], str]) -> None:
        super().__init__()
        self.cls = dict
        self.ftype = FType.SHAPE
        self.raw_types = raw_shape_types
        self.exact_type = False

    def define(self, fdef: Fdef) -> None:
        super().define(fdef)
        fdef._raw_shape_types = self.raw_types

    def validate(self, ctx: Ctx) -> None:
        if ctx.val is None:
            return
        super().validate(ctx)
        all_fields = ctx.ctxcfg.all_fields
        if all_fields is None:
            all_fields = ctx.cdefowner.jconf.validate_all_fields
        ctor = VMsgCollector()
        for k, types in ctx.fdef.shape_types.items():
            try:
                value_at_key = ctx.val[k]
            except KeyError:
                value_at_key = None
            try:
                ival = ctx.colval(value_at_key, k, types.fdef, ctx.val)
                types.modifier.validate(ival)
            except ValidationException as exception:
                if all_fields:
                    ctor.receive(exception.keypath_messages)
                else:
                    raise exception
        if ctor.has_msgs:
            ctx.raise_mvexc(ctor.messages)

    def _has_field_value(self,
                         field_key: str,
                         keys: Sequence[str],
                         jconf: JConf) -> bool:
        field_name = field_key
        json_field_name = jconf.key_encoding_strategy(field_name)
        return json_field_name in keys or field_name in keys

    def _get_field_value(self,
                         field_key: str,
                         value: dict[str, Any],
                         jconf: JConf) -> Any:
        field_value = value.get(field_key)
        if field_value is None:
            field_value = value.get(jconf.key_encoding_strategy(field_key))
        return field_value

    def _strictness_check(self, value: dict[str, Any], ctx: Ctx):
        value_keys = list(value.keys())
        value_keys = [ctx.jconfowner.key_decoding_strategy(k) for k in value_keys]
        keys = ctx.fdef.shape_types.keys()
        for k in value_keys:
            if k not in keys:
                kp = concat_keypath(ctx.skeypathr, k)
                ctx.raise_mvexc({kp: 'key is not allowed'})

    def transform(self, ctx: Ctx) -> Any:
        value = ctx.val
        fd = ctx.fdef
        if fd.collection_nullability == Nullability.NONNULL and value is None:
            value = {}
        if value is None:
            return None
        if not isinstance(value, dict):
            return value
        if fd.strictness == Strictness.STRICT:
            self._strictness_check(value, ctx)
        retval = {}
        for fk, types in ctx.fdef.shape_types.items():
            fv = None
            if self._has_field_value(fk, list(value.keys()), ctx.jconfowner):
                fv = self._get_field_value(fk, value, ctx.jconfowner)
            ictx = ctx.colval(fv, fk, types.fdef, value)
            retval[fk] = types.modifier.transform(ictx)
        return retval

    def tojson(self, ctx: Ctx) -> Any:
        if ctx.val is None:
            return None
        if not isinstance(ctx.val, dict):
            return ctx.val
        retval = {}
        for k, types in ctx.fdef.shape_types.items():
            key = ctx.jconfowner.key_encoding_strategy(k)
            value_at_key = ctx.val.get(k)
            if types:
                retval[key] = types.modifier.tojson(ctx.nval(value_at_key))
            else:
                retval[key] = value_at_key
        return retval

    def serialize(self, ctx: Ctx) -> Any:
        if ctx.val is None:
            return None
        if not isinstance(ctx.val, dict):
            return ctx.val
        retval = {}
        for key, types in ctx.fdef.shape_types.items():
            value_at_key = ctx.val.get(key)
            if types:
                ictx = ctx.colval(value_at_key, key, types.fdef, ctx.val)
                retval[key] = types.modifier.serialize(ictx)
            else:
                retval[key] = value_at_key
        return retval



"""module for operator modifier."""
from __future__ import annotations
from typing import Callable, TYPE_CHECKING
from inspect import signature
from ..excs import ValidationException
from .modifier import Modifier
if TYPE_CHECKING:
    from ..ctx import Ctx


class OpModifier(Modifier):
    """Op modifier validates with user defined modifier."""

    def __init__(self, op_callable: Callable) -> None:
        if not callable(op_callable):
            raise ValueError('op modifier is not callable')
        params_len = len(signature(op_callable).parameters)
        if params_len > 4 or params_len < 1:
            raise ValueError('not a valid op modifier')
        self.op_callable = op_callable

    def validate(self, ctx: Ctx) -> None:
        if ctx.operator is None:
            ctx.raise_vexc('operator not present')
        params_len = len(signature(self.op_callable).parameters)
        if params_len == 1:
            result = self.op_callable(ctx.operator)
        elif params_len == 2:
            result = self.op_callable(ctx.operator, ctx.owner)
        elif params_len == 3:
            result = self.op_callable(ctx.operator, ctx.owner, ctx.val)
        elif params_len == 4:
            result = self.op_callable(ctx.operator, ctx.owner, ctx.val, ctx)
        if result is None:
            return
        if result is True:
            return
        if result is False:
            ctx.raise_vexc('unauthorized operation')
        if isinstance(result, str):
            ctx.raise_vexc(result)
        raise ValueError('invalid modifier')



"""module for minlength modifier."""
from __future__ import annotations
from typing import TYPE_CHECKING
from ..excs import ValidationException
from .modifier import Modifier
if TYPE_CHECKING:
    from ..ctx import Ctx


class MinlengthModifier(Modifier):
    """Minlength modifier validates value against min length."""

    def __init__(self, minlength: int) -> None:
        self.minlength = minlength

    def validate(self, ctx: Ctx) -> None:
        if ctx.val is None:
            return
        if len(ctx.val) < self.minlength:
            ctx.raise_vexc('length of value is not greater than or equal '
                           f'{self.minlength}')



from __future__ import annotations
from typing import Optional
from jsonclasses import jsonclass, types


@jsonclass
class SimpleConfigUser:
    email: Optional[str] = types.str
    config: dict[str, bool] = types.shape({
        'ios': types.bool.required,
        'android': types.bool.required
    }).required



"""module for length modifier."""
from __future__ import annotations
from typing import Optional, TYPE_CHECKING
from ..excs import ValidationException
from .modifier import Modifier
if TYPE_CHECKING:
    from ..ctx import Ctx


class LengthModifier(Modifier):
    """Length modifier validate value against the provided length."""

    def __init__(self, minlength: int, maxlength: Optional[int]) -> None:
        self.minlength = minlength
        self.maxlength = maxlength if maxlength is not None else minlength

    def validate(self, ctx: Ctx) -> None:
        if ctx.val is None:
            return
        value = ctx.val
        if len(value) > self.maxlength or len(value) < self.minlength:
            if self.minlength != self.maxlength:
                message = f'length of value should be in between {self.minlength} and {self.maxlength}'
            else:
                message = f'length of value is not {self.minlength}'
            ctx.raise_vexc(message)



"""module for linkto modifier."""
from ..fdef import DeleteRule, Fdef, FStore
from .modifier import Modifier


class LinkToModifier(Modifier):
    """Link to modifier marks a field which is a local key."""

    def define(self, fdef: Fdef) -> None:
        fdef._fstore = FStore.LOCAL_KEY
        if fdef._delete_rule is None:
            fdef._delete_rule = DeleteRule.NULLIFY



from __future__ import annotations
from typing import TypedDict
from jsonclasses import jsonclass, types


class EmailSetting(TypedDict):
    auto_send: bool
    receive_promotion: bool


@jsonclass
class SimpleSetting:
    user: str
    email: EmailSetting = types.shape({
        'auto_send': types.bool.required,
        'receive_promotion': types.bool.required
    })



"""module for match modifier."""
from __future__ import annotations
from typing import TYPE_CHECKING
from re import search
from ..excs import ValidationException
from .modifier import Modifier
if TYPE_CHECKING:
    from ..ctx import Ctx


class MatchModifier(Modifier):
    """Match modifier validates value against pattern."""

    def __init__(self, pattern: str) -> None:
        self.pattern = pattern

    def validate(self, ctx: Ctx) -> None:
        if isinstance(ctx.val, str) and search(self.pattern, ctx.val) is None:
            ctx.raise_vexc(f'value does not match \'{self.pattern}\'')



"""module for inputlname modifier."""
from ..fdef import Fdef, EnumInput
from .modifier import Modifier


class InputLnameModifier(Modifier):
    """Input lname modifier tweaks enum modifier's behavior."""

    def define(self, fdef: Fdef) -> None:
        if fdef.enum_input is None:
            fdef._enum_input = EnumInput.LOWERCASE_NAME
        else:
            fdef._enum_input = fdef.enum_input | EnumInput.LOWERCASE_NAME
        return



"""module for index modifier."""
from .modifier import Modifier
from ..fdef import Fdef


class IndexModifier(Modifier):
    """Index modifier implies this column should be indexed in database."""

    def define(self, fdef: Fdef) -> None:
        fdef._index = True



from __future__ import annotations
from jsonclasses import jsonclass, types
from jsonclasses.keypath import camelize_key, identical_key, underscore_key


@jsonclass(key_encoding_strategy=camelize_key,
           key_decoding_strategy=underscore_key)
class CamelizedWeather:
    data: dict[str, str] = types.dictof(str)


@jsonclass(key_encoding_strategy=identical_key,
           key_decoding_strategy=identical_key)
class UncamelizedWeather:
    data: dict[str, str] = types.dictof(str)



from __future__ import annotations
from jsonclasses import jsonclass, types


@jsonclass
class OpUser:
    name: str = types.str.required
    owned_teams: list[OpTeam] = types.nonnull.listof('OpTeam') \
                                     .linkedby('owner')


@jsonclass
class OpTeam:
    name: str = types.str.op(lambda o, t: o == t.owner).required
    owner: OpUser = types.instanceof('OpUser').linkto.required


@jsonclass
class AsopUser:
    name: str = types.str.required
    owned_teams: list[AsopTeam] = types.nonnull.listof('AsopTeam') \
                                       .linkedby('owner')


@jsonclass
class AsopTeam:
    name: str = types.str.required
    owner: AsopUser = types.instanceof('AsopUser').linkto.asop(lambda o: o)


@jsonclass
class AsopdUser:
    name: str = types.str.required
    owned_teams: list[AsopdTeam] = types.nonnull.listof('AsopdTeam') \
                                        .linkedby('owner')


@jsonclass
class AsopdTeam:
    name: str = types.str.required
    owner: AsopdUser = types.instanceof('AsopdUser').linkto.asopd



from __future__ import annotations
from typing import Optional, TypedDict
from jsonclasses import jsonclass, types


class EmailSetting(TypedDict):
    auto_send: bool
    receive_promotion: bool


@jsonclass
class SimpleShorthandSetting:
    user: Optional[str]
    email: EmailSetting = types.nonnull.shape({
        'auto_send': Optional[bool],
        'receive_promotion': Optional[bool]
    })



from __future__ import annotations
from jsonclasses.keypath import camelize_key, underscore_key
from typing import Optional
from jsonclasses import jsonclass, types


@jsonclass(key_encoding_strategy=camelize_key,
           key_decoding_strategy=underscore_key)
class SimpleFolder:
    name: Optional[str] = types.str
    config: dict[str, bool] = types.strict.shape({
        'display_size': types.bool.required,
        'display_date': types.bool.required
    }).required



"""module for max modifier."""
from __future__ import annotations
from typing import Union, TYPE_CHECKING
from .modifier import Modifier
if TYPE_CHECKING:
    from ..ctx import Ctx


class MaxModifier(Modifier):
    """Max modifier validates value against max value."""

    def __init__(self, max_value: Union[int, float]) -> None:
        self.max_value = max_value

    def validate(self, ctx: Ctx) -> None:
        is_number = type(ctx.val) is int or type(ctx.val) is float
        if is_number and ctx.val > self.max_value:
            ctx.raise_vexc(f'value is not less than or equal {self.max_value}')



"""module for range modifier."""
from __future__ import annotations
from typing import Union, TYPE_CHECKING
from .modifier import Modifier
from .min_modifier import MinModifier
from .max_modifier import MaxModifier
if TYPE_CHECKING:
    from ..ctx import Ctx


class RangeModifier(Modifier):
    """A range modifier validates value against a range."""

    def __init__(self, min: Union[int, float], max: Union[int, float]):
        self.min = min
        self.max = max

    def validate(self, ctx: Ctx) -> None:
        if ctx.val is None:
            return
        MinModifier(self.min).validate(ctx)
        MaxModifier(self.max).validate(ctx)



from __future__ import annotations
from typing import Optional, TypedDict
from jsonclasses import jsonclass, jsondict


@jsondict
class Preference(TypedDict):
    ios: bool
    android: bool


@jsonclass
class AutoStrSetting:
    user: Optional[str]
    preference: Preference



from typing import TypedDict
from jsonclasses import jsonclass, types


class Settings(TypedDict):
    ios: bool
    android: bool
    name: str


@jsonclass
class DefaultShape:
    settings: Settings = types.nonnull.shape({
        'ios': types.bool.default(True).required,
        'android': types.bool.default(True).required,
        'name': types.str.default('unnamed').required
    })



"""
This module contains `jsonclass`, the decorator for JSON Classes.
"""
from jsonclasses.keypath import identical_key
from typing import Optional, Union, Callable, overload, cast
from dataclasses import dataclass
from .jconf import (JConf, OnCreate, CanCreate, OnDelete, CanUpdate,
                     CanDelete, CanRead, OnUpdate)
from .jfield import JField
from .cdef import Cdef
from .jsonclassify import jsonclassify
from .jobject import JObject


@overload
def jsonclass(cls: type) -> type[JObject]: ...


@overload
def jsonclass(
    cls: None = None,
    class_graph: Optional[str] = 'default',
    key_encoding_strategy: Optional[Callable[[str], str]] = None,
    key_decoding_strategy: Optional[Callable[[str], str]] = None,
    camelize_json_keys: Optional[bool] = None,
    strict_input: Optional[bool] = None,
    ref_key_encoding_strategy: Optional[Callable[[JField], str]] = None,
    validate_all_fields: Optional[bool] = None,
    abstract: Optional[bool] = None,
    reset_all_fields: Optional[bool] = None,
    on_create: Optional[Union[OnCreate, list[OnCreate]]] = None,
    on_update: Optional[Union[OnUpdate, list[OnUpdate]]] = None,
    on_delete: Optional[Union[OnDelete, list[OnDelete]]] = None,
    can_create: Optional[Union[CanCreate, list[CanCreate]]] = None,
    can_update: Optional[Union[CanUpdate, list[CanUpdate]]] = None,
    can_delete: Optional[Union[CanDelete, list[CanDelete]]] = None,
    can_read: Optional[Union[CanRead, list[CanRead]]] = None,
) -> Callable[[type], type[JObject]]: ...


@overload
def jsonclass(
    cls: type,
    class_graph: Optional[str] = 'default',
    key_encoding_strategy: Optional[Callable[[str], str]] = None,
    key_decoding_strategy: Optional[Callable[[str], str]] = None,
    camelize_json_keys: Optional[bool] = None,
    strict_input: Optional[bool] = None,
    ref_key_encoding_strategy: Optional[Callable[[JField], str]] = None,
    validate_all_fields: Optional[bool] = None,
    abstract: Optional[bool] = None,
    reset_all_fields: Optional[bool] = None,
    on_create: Optional[Union[OnCreate, list[OnCreate]]] = None,
    on_update: Optional[Union[OnUpdate, list[OnUpdate]]] = None,
    on_delete: Optional[Union[OnDelete, list[OnDelete]]] = None,
    can_create: Optional[Union[CanCreate, list[CanCreate]]] = None,
    can_update: Optional[Union[CanUpdate, list[CanUpdate]]] = None,
    can_delete: Optional[Union[CanDelete, list[CanDelete]]] = None,
    can_read: Optional[Union[CanRead, list[CanRead]]] = None,
) -> type[JObject]: ...


def jsonclass(
    cls: Optional[type] = None,
    class_graph: Optional[str] = 'default',
    key_encoding_strategy: Optional[Callable[[str], str]] = None,
    key_decoding_strategy: Optional[Callable[[str], str]] = None,
    camelize_json_keys: Optional[bool] = None,
    strict_input: Optional[bool] = None,
    ref_key_encoding_strategy: Optional[Callable[[JField], str]] = None,
    validate_all_fields: Optional[bool] = None,
    abstract: Optional[bool] = None,
    reset_all_fields: Optional[bool] = None,
    on_create: Optional[Union[OnCreate, list[OnCreate]]] = None,
    on_update: Optional[Union[OnUpdate, list[OnUpdate]]] = None,
    on_delete: Optional[Union[OnDelete, list[OnDelete]]] = None,
    can_create: Optional[Union[CanCreate, list[CanCreate]]] = None,
    can_update: Optional[Union[CanUpdate, list[CanUpdate]]] = None,
    can_delete: Optional[Union[CanDelete, list[CanDelete]]] = None,
    can_read: Optional[Union[CanRead, list[CanRead]]] = None,
) -> Union[Callable[[type], type[JObject]], type[JObject]]:
    """The jsonclass object class decorator. To declare a jsonclass class, use
    this syntax:

      @jsonclass
      class MyObject:
        my_field_one: str
        my_field_two: bool
    """
    if cls is not None:
        if not isinstance(cls, type):
            raise ValueError('@jsonclass should be used to decorate a class.')
        if camelize_json_keys is False:
            key_encoding_strategy = identical_key
            key_decoding_strategy = identical_key
        jconf = JConf(
            cgraph=cast(str, class_graph),
            key_encoding_strategy=key_encoding_strategy,
            key_decoding_strategy=key_decoding_strategy,
            strict_input=strict_input,
            ref_key_encoding_strategy=ref_key_encoding_strategy,
            validate_all_fields=validate_all_fields,
            abstract=abstract,
            reset_all_fields=reset_all_fields,
            on_create=on_create,
            on_update=on_update,
            on_delete=on_delete,
            can_create=can_create,
            can_update=can_update,
            can_delete=can_delete,
            can_read=can_read)
        dcls: type = dataclass(init=False)(cls)
        jcls = jsonclassify(dcls)
        cdef = Cdef(jcls, jconf)
        jcls.cdef = cdef
        jconf.cgraph.put(cdef)
        return jcls
    else:
        def parametered_jsonclass(cls):
            return jsonclass(
                cls,
                class_graph=class_graph,
                key_encoding_strategy=key_encoding_strategy,
                key_decoding_strategy=key_decoding_strategy,
                camelize_json_keys=camelize_json_keys,
                strict_input=strict_input,
                ref_key_encoding_strategy=ref_key_encoding_strategy,
                validate_all_fields=validate_all_fields,
                abstract=abstract,
                reset_all_fields=reset_all_fields,
                on_create=on_create,
                on_update=on_update,
                on_delete=on_delete,
                can_create=can_create,
                can_update=can_update,
                can_delete=can_delete,
                can_read=can_read)
        return parametered_jsonclass



"""This module contains all modifier markers."""
# flake8: noqa: F401
from .modifier import Modifier
from .invalid_modifier import InvalidModifier

from .primary_modifier import PrimaryModifier

# access markers
from .writeonly_modifier import WriteonlyModifier
from .writeonce_modifier import WriteonceModifier
from .writenonnull_modifier import WriteNonnullModifier
from .readonly_modifier import ReadonlyModifier
from .readwrite_modifier import ReadwriteModifier
from .temp_modifier import TempModifier

# database index command markers
from .index_modifier import IndexModifier
from .unique_modifier import UniqueModifier

# orm relationship command markers

from .embedded_modifier import EmbeddedModifier
from .linkto_modifier import LinkToModifier
from .linkedby_modifier import LinkedByModifier
from .linkedthru_modifier import LinkedThruModifier
from .linkedin_modifier import LinkedInModifier
from .referrer_modifier import ReferrerModifier
from .referee_modifier import RefereeModifier
from .deny_modifier import DenyModifier
from .cascade_modifier import CascadeModifier
from .nullify_modifier import NullifyModifier

# eager validation markers
from .eager_modifier import EagerModifier

# preserialize validation markers
from .preserialize_modifier import PreserializeModifier

# str modifiers
from .str_modifier import StrModifier
from .match_modifier import MatchModifier
from .oneof_modifier import OneOfModifier
from .truncate_modifier import TruncateModifier
from .trim_modifier import TrimModifier
from .minlength_modifier import MinlengthModifier
from .maxlength_modifier import MaxlengthModifier
from .length_modifier import LengthModifier
from .url_modifier import UrlModifier
from .email_modifier import EmailModifier
from .digit_modifier import DigitModifier
from .alpha_modifier import AlphaModifier
from .numeric_modifier import NumericModifier
from .alnum_modifier import AlnumModifier
from .totitle_modifier import ToTitleModifier
from .tocap_modifier import ToCapModifier
from .tolower_modifier import ToLowerModifier
from .toupper_modifier import ToUpperModifier

# number modifiers
from .int_modifier import IntModifier
from .float_modifier import FloatModifier
from .min_modifier import MinModifier
from .max_modifier import MaxModifier
from .lt_modifier import LtModifier
from .gt_modifier import GtModifier
from .nonnegative_modifier import NonnegativeModifier
from .nonpositive_modifier import NonpositiveModifier
from .range_modifier import RangeModifier
from .positive_modifier import PositiveModifier
from .negative_modifier import NegativeModifier
from .round_modifier import RoundModifier

# bool modifiers
from .bool_modifier import BoolModifier

# datetime modifiers
from .date_modifier import DateModifier
from .datetime_modifier import DatetimeModifier

# enum modifiers
from .enum_modifier import EnumModifier
from .inputvalue_modifier import InputValueModifier
from .inputname_modifier import InputNameModifier
from .inputlname_modifier import InputLnameModifier
from .inputall_modifier import InputAllModifier
from .outputvalue_modifier import OutputValueModifier
from .outputname_modifier import OutputNameModifier
from .outputlname_modifier import OutputLnameModifier

# collection modifiers
from .listof_modifier import ListOfModifier
from .dictof_modifier import DictOfModifier

# object modifiers
from .shape_modifier import ShapeModifier
from .strict_modifier import StrictModifier
from .instanceof_modifier import InstanceOfModifier

# mixed type modifier
from .union_modifier import UnionModifier
from .any_modifier import AnyModifier

# nullability modifiers
from .required_modifier import RequiredModifier
from .nullable_modifier import NullableModifier
from .present_modifier import PresentModifier
from .presentwith_modifier import PresentWithModifier
from .presentwithout_modifier import PresentWithoutModifier

# custom modifier
from .validate_modifier import ValidateModifier

# operator modifier
from .op_modifier import OpModifier

# comparing modifier and callback
from .reset_modifier import ResetModifier
from .compare_modifier import CompareModifier

# default transformer
from .default_modifier import DefaultModifier

# transform
from .transform_modifier import TransformModifier

# operator transform
from .asop_modifier import AsopModifier
from .asopd_modifier import AsopdModifier

# shape transformer
from .nonnull_modifier import NonnullModifier

# chained modifier
from .chained_modifier import ChainedModifier

# setonsave setter, onsave callback
from .setonsave_modifier import SetOnSaveModifier
from .onsave_modifier import OnSaveModifier
from .onupdate_modifier import OnUpdateModifier
from .onwrite_modifier import OnWriteModifier

# jsonclasses internal
from .unresolved_modifier import UnresolvedModifier



"""module for validate modifier."""
from __future__ import annotations
from typing import Callable, TYPE_CHECKING
from inspect import signature
from ..excs import ValidationException
from .modifier import Modifier
if TYPE_CHECKING:
    from ..ctx import Ctx


class ValidateModifier(Modifier):
    """Validate modifier takes a modifier."""

    def __init__(self, validate_callable: Callable) -> None:
        if not callable(validate_callable):
            raise ValueError('modifier is not callable')
        params_len = len(signature(validate_callable).parameters)
        if params_len > 2 or params_len < 1:
            raise ValueError('not a valid modifier')
        self.validate_callable = validate_callable

    def validate(self, ctx: Ctx) -> None:
        if ctx.val is None:
            return
        params_len = len(signature(self.validate_callable).parameters)
        if params_len == 1:
            result = self.validate_callable(ctx.val)
        elif params_len == 2:
            result = self.validate_callable(ctx.val, ctx)
        if result is None:
            return
        if result is True:
            return
        if result is False:
            ctx.raise_vexc('invalid value')
        if isinstance(result, str):
            ctx.raise_vexc(result)
        raise ValueError('invalid modifier')



from __future__ import annotations
from typing import Optional
from jsonclasses import jsonclass, types
from jsonclasses.keypath import identical_key


@jsonclass(key_encoding_strategy=identical_key,
           key_decoding_strategy=identical_key)
class SimpleNode:
    name: Optional[str] = types.str
    config: dict[str, bool] = types.shape({
        'display_size': types.bool.required,
        'display_date': types.bool.required
    }).required



"""module for oneoftype modifier."""
from __future__ import annotations
from typing import TYPE_CHECKING, Any
from ..excs import ValidationException
from .modifier import Modifier
from ..rtypes import rtypes
from ..fdef import Fdef, FType
if TYPE_CHECKING:
    from ..ctx import Ctx


class UnionModifier(Modifier):
    """Union type modifier validates value against a list of available types.
    """

    def __init__(self, type_list: list[Any]) -> None:
        self.type_list = type_list

    def define(self, fdef: Fdef) -> None:
        fdef._ftype = FType.UNION
        fdef._raw_union_types = [rtypes(t) for t in self.type_list]

    def validate(self, ctx: Ctx) -> None:
        if ctx.val is None:
            return
        for types in ctx.fdef.raw_union_types:
            try:
                types.modifier.validate(ctx)
                return
            except ValidationException:
                continue
        ctx.raise_vexc('value is not of any provided type')



from __future__ import annotations
from typing import Optional, Union
from jsonclasses import jsonclass, types


@jsonclass
class SimpleMixed:
    name: Optional[str]
    mixed: Optional[Union[int, str]] = types.union([str, int])
